/**
 * Author: Hendrik van Arragon, 2022
 * SPDX-License-Identifier: MIT
 */

#include <sched.h>
#include <unistd.h>
#include <time.h>

#include <functional>
#include <memory>
#include <map>
#include <iostream>
#include <fstream>
#include <filesystem>
#include <clocale>

#include <abmt/task.h>
#include <abmt/time.h>
#include <abmt/rt/com.h>
#include <abmt/rt/raster.h>
#include <abmt/serialize_utils.h>
#include <abmt/mutex.h>
#include <abmt/os.h>
#include <abmt/util/bt_exception.h>
#include <abmt/json.h>

#include "model_adapter_std.h"
#include "model_com_eio.h"

using namespace std;
using json = abmt::json;

volatile bool global_stop = false;

model_com_eio* com_ptr = 0;

void on_sig(int sig){
	global_stop = true;
}

void on_sigsegv(int sig){
	static abmt::mutex lock;
	auto l = lock.get_scope_lock();
	abmt::util::bt_exception ex("");
	cout << "Segmentation Fault:"<< endl;
	ex.print_backtrace();
	global_stop = true;
	cout.flush();
	exit(1);
}

void abmt::log(std::string s){
	if(com_ptr->tcp_con->connected){
		com_ptr->adapter.log(s);
	}else{
		cout << s << endl;
	}
}
void abmt::log_err(std::string s){
	if(com_ptr->tcp_con->connected){
		com_ptr->adapter.log_err(s);
	}else{
		cout << s << endl;
	}
}

void abmt::die(std::string s){
	throw abmt::util::bt_exception(s);
}

void abmt::die_if(bool condition, std::string msg){
	if(condition){
		abmt::die(msg);
	}
}

abmt::time abmt::time::now(){
	timespec t;
	clock_gettime(CLOCK_REALTIME, &t);
	int64_t res;
	res = t.tv_sec * 1000LL*1000*1000 + t.tv_nsec; // 1000LL very important for 32bit systems
	return time(res);
}

class live_task: public abmt::task{
public:
	abmt::rt::raster* r;
	uint32_t raster_index;
	abmt::time start_time;

	bool stop = false;

	live_task(abmt::rt::raster* r, uint32_t r_idx):abmt::task([this]{this->run();},false), r(r), raster_index(r_idx){

	}

	void run(){
		try{
			r->init();
			usleep(10*1000); // give async tasks some time to init (see example/io/udp.acfg)
			start_time = abmt::time::now();
			if(r->is_sync == false){
				while(true){
					auto time_to_wait = r->poll();
					if( time_to_wait > 0 ){
						usleep(time_to_wait.us());
					}else{
						break;
					}
					if(global_stop){
						return; // no init -> no final;
					}
				}
			}
			r->init_tick();
			com_ptr->adapter.send_daq(raster_index);
			r->n_ticks++;
			while(stop == false && global_stop == false){
				if(r->is_sync){
					auto t_next_tick = start_time + r->n_ticks*r->interval;
					abmt::time delay =  t_next_tick - abmt::time::now();

					if(delay >= 0){
						if(delay > r->interval){
							// maybe someone messed around with the clock (ntpd?)
							delay = r->interval;
							r->n_ticks = 0;
							start_time = abmt::time::now();
						}
						usleep(delay.us());
						if(global_stop){
							break;
						}
					}else{
						r->n_ticks = 0;
						start_time = abmt::time::now();
					}
				}else{
					auto time_to_wait = r->poll();
					if( time_to_wait > 0 ){
						usleep(time_to_wait.us());
						continue;
					}
				}
				r->tick();
				com_ptr->adapter.send_daq(raster_index);
				r->n_ticks++;

			}
			r->final();
		}catch(abmt::util::bt_exception& b){
			cout << "An error occured in task " << r->name << ": " << b.what() << endl;
			b.print_backtrace();
			global_stop = true;
		}
	}

};
int main(int argc, char* argv[]){

	/*
	cpu_set_t  mask;
	CPU_ZERO(&mask);
	CPU_SET(2, &mask);
	int result = sched_setaffinity(0, sizeof(mask), &mask);
	if(result == -1){
		std::cout << "unable to set sched_setaffinit for process. This could cause realtime-issues. Error-message: "<< std::strerror(errno) << std::endl;
	}
 	*/

	signal(SIGHUP,on_sig);
	signal(SIGINT,on_sig);
	signal(SIGQUIT,on_sig);
	signal(SIGSEGV,on_sigsegv);


	std::setlocale(LC_ALL, "C.UTF-8");

	std::string param_json  = "parameters.json";

	if(argc > 1){
		if(std::string(argv[1]) == "--help"){
			cout << "usage: " << argv[0] << " <parameters.json>" << endl;
			return 0;
		}
		param_json = argv[1];
	}

	auto model_ptr = abmt::rt::get_model();

	abmt::io::event_list e;
	model_com_eio mdl_com(e);
	com_ptr = &mdl_com;
	mdl_com.set_model(model_ptr);
	mdl_com.adapter.on_quit = [&](){
		global_stop = true;
	};
	mdl_com.adapter.on_save_parameters = [&](){
		abmt::json res;
		for(auto p: model_ptr->parameters){
			std::string key = p.name;
			res[p.name] = abmt::serialize::to_json(p.parameter->get_type());
		}
		ofstream ofs(param_json);
		ofs << res.dump();
		ofs.close();
	};

	if(std::filesystem::exists(param_json)){
		ifstream f(param_json);
		stringstream ss;
		ss << f.rdbuf();
		f.close();
		json params= json::parse(ss.str());
		for(auto p: params){
			for(auto mdl_param: model_ptr->parameters){
				std::string name(mdl_param.name);
				if(name == p.key){
					auto exp = mdl_param.parameter->get_type();
					try{
						abmt::serialize::from_json(p.value,exp);
					}catch(...){
						std::cout << "error parsing parameter: " << name << std::endl;
					}
					break;
				}
			}
		}
	}
	mdl_com.adapter.on_save_parameters(); // update parameters for reproducibility

	std::map<std::string, std::shared_ptr<live_task>> raster_tasks;

	for(size_t i = 0; i < model_ptr->rasters.length; ++i){
		auto r = model_ptr->rasters[i];

		if(r){
			auto task = make_shared<live_task>(r,i);
			raster_tasks[r->name] = task;
			task->start();
		}
	}

	try{
		while(global_stop == false){
			e.wait(100);
		}
		// quit
		bool all_tasks_ready = false;
		bool force_quit = false;
		abmt::io::timer quit_timer(e, 5000, [&](){ force_quit = true; });
		for(auto t: raster_tasks){
			if(t.second->is_running()){
				t.second->send_sig(SIGINT); // stops blocking actions in raster
			}
		}
		while(all_tasks_ready == false && force_quit == false){
			all_tasks_ready = true;
			for(auto t: raster_tasks){
				if(t.second->is_running()){
					all_tasks_ready = false;
					break;
				}
			}
			e.wait(100);
		}

		if(force_quit){
			cout << "Force exit. The following rasters are still running:" << endl;
			for(auto t: raster_tasks){
				if(t.second->is_running()){
					cout << t.first << endl;
				}
			}
			return 1;
		}

	}catch(abmt::util::bt_exception& b){
		cout << "An error occured in the main task: " << b.what() << endl;
		b.print_backtrace();
		global_stop = true;
	}
	return 0;
}
